#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime, timedelta, timezone
from typing import List

from fabric_mb.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric_mb.message_bus.messages.reservation_predecessor_avro import ReservationPredecessorAvro
from fim.slivers.base_sliver import BaseSliver

from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.util.id import ID


class ReservationConverter:
    """
    Class responsible for computing reservations from slivers
    """
    def __init__(self, *, controller: ABCMgmtControllerMixin, broker: ID):
        self.controller = controller
        self.broker = broker

    def generate_reservation(self, *, sliver: BaseSliver, slice_id: str, end_time: datetime,
                             pred_list: List[str] = None) -> LeaseReservationAvro:
        """
        Responsible to generate reservation from the sliver
        :param sliver Network Service or Network Node Sliver
        :param slice_id Slice Id
        :param end_time End Time
        :param pred_list Predecessor Reservation Id List
        :returns list of tickets
        """
        ticket = LeaseReservationAvro()
        ticket.redeem_processors = []
        if pred_list is not None:
            for rid in pred_list:
                pred = ReservationPredecessorAvro()
                pred.set_reservation_id(value=rid)
                ticket.redeem_processors.append(pred)
        ticket.set_slice_id(slice_id)
        ticket.set_broker(str(self.broker))
        ticket.set_units(1)
        ticket.set_resource_type(str(sliver.get_type()))
        start = datetime.now(timezone.utc)
        end = start + timedelta(hours=Constants.DEFAULT_LEASE_IN_HOURS)
        if end_time is not None:
            end = end_time
        ticket.set_start(ActorClock.to_milliseconds(when=start))
        ticket.set_end(ActorClock.to_milliseconds(when=end))
        ticket.set_state(ReservationStates.Unknown.value)
        ticket.set_pending_state(ReservationPendingStates.None_.value)
        ticket.set_sliver(sliver)
        ticket.set_reservation_id(value=str(ID()))
        return ticket




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time
import traceback
from typing import List, Dict

from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.poa_info_avro import PoaInfoAvro

from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
from fabric_cf.actor.core.apis.abc_timer_task import ABCTimerTask
from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin, ActorType
from fabric_cf.actor.core.apis.abc_actor_event import ABCActorEvent
from fabric_cf.actor.core.apis.abc_actor_proxy import ABCActorProxy
from fabric_cf.actor.core.apis.abc_actor_runnable import ABCActorRunnable
from fabric_cf.actor.core.apis.abc_query_response_handler import ABCQueryResponseHandler
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.common.exceptions import ActorException
from fabric_cf.actor.core.container.message_service import MessageService
from fabric_cf.actor.core.core.event_processor import TickEvent, EventType, EventProcessor
from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC
from fabric_cf.actor.core.kernel.kernel_wrapper import KernelWrapper
from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.slice import SliceTypes
from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
from fabric_cf.actor.core.proxies.proxy import Proxy
from fabric_cf.actor.core.container.maintenance import Site
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.reflection_utils import ReflectionUtils
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.security.auth_token import AuthToken


class ActorMixin(ABCActorMixin):
    """
    Actor is the base class for all actor implementations
    """
    DefaultDescription = "no description"
    SUPPORTED_EVENTS = [EventType.TickEvent, EventType.InterActorEvent, EventType.SyncEvent]

    def __init__(self, *, auth: AuthToken = None, clock: ActorClock = None):
        # Globally unique identifier for this actor.
        self.guid = ID()
        # Actor name.
        self.name = None
        # Actor type code.
        self.type = ActorType.All
        # Actor description.
        self.description = self.DefaultDescription
        # Identity object representing this actor.
        self.identity = auth
        # Actor policy object.
        self.policy = None
        # Actor plugin
        self.plugin = None
        # True if this actor has completed the recovery phase.
        self.recovered = False
        # The kernel wrapper.
        self.wrapper = None
        # logger
        self.logger = None
        # Factory for term.
        self.clock = clock
        # current cycle
        self.current_cycle = -1
        # True if the current tick is the first tick this actor has received.
        self.first_tick = True
        # Set to true when the actor is stopped.
        self.stopped = False
        # Initialization status.
        self.initialized = False
        # Reservations to close once recovery is complete.
        self.closing = ReservationSet()

        self.message_service = None
        self.event_processors = {}

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['recovered']
        del state['wrapper']
        del state['logger']
        del state['clock']
        del state['current_cycle']
        del state['first_tick']
        del state['stopped']
        del state['initialized']
        del state['closing']
        del state['message_service']
        del state['event_processors']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.recovered = False
        self.wrapper = None
        self.logger = None
        self.clock = None
        self.current_cycle = -1
        self.first_tick = True
        self.stopped = False
        self.initialized = False
        self.closing = ReservationSet()
        self.message_service = None
        self.policy.set_actor(actor=self)
        self.event_processors = {}

    def actor_added(self, *, config: ActorConfig):
        self.plugin.actor_added(config=config)

    def actor_removed(self):
        return

    def fail(self, *, rid: ID, message: str):
        self.wrapper.fail(rid=rid, message=message)

    def fail_delegation(self, *, did: str, message: str):
        self.wrapper.fail_delegation(did=did, message=message)

    def close_delegation(self, *, did: str):
        self.wrapper.close_delegation(did=did)

    def close_by_rid(self, *, rid: ID):
        self.wrapper.close(rid=rid)

    def close(self, *, reservation: ABCReservationMixin):
        if reservation is not None:
            if not self.recovered:
                #self.logger.debug("Adding reservation: {} to closing list".format(reservation.get_reservation_id()))
                self.closing.add(reservation=reservation)
            else:
                #self.logger.debug("Closing reservation: {}".format(reservation.get_reservation_id()))
                self.wrapper.close(rid=reservation.get_reservation_id())

    def close_slice_reservations(self, *, slice_id: ID):
        self.wrapper.close_slice_reservations(slice_id=slice_id)

    def close_reservations(self, *, reservations: ReservationSet):
        for reservation in reservations.reservations.values():
            try:
                #self.logger.debug("Closing reservation: {}".format(reservation.get_reservation_id()))
                self.close(reservation=reservation)
            except Exception as e:
                self.logger.error(traceback.format_exc())
                self.logger.error("Could not close for #{} {}".format(reservation.get_reservation_id(), e))

    def error(self, *, err: str):
        """
        Logs and propagates a general error.

        @param err
                   log/exception message.
        @throws Exception
                    always
        """
        self.logger.error(err)
        raise ActorException(err)

    def extend(self, *, rid: ID, resources: ResourceSet, term: Term, dependencies: List[ABCReservationMixin] = None):
        self.wrapper.extend_reservation(rid=rid, resources=resources, term=term, dependencies=dependencies)

    def external_tick(self, *, cycle: int):
        self.logger.info("External Tick start cycle: {}".format(cycle))
        self.queue_event(incoming=TickEvent(actor=self, cycle=cycle))
        self.logger.info("External Tick end cycle: {}".format(cycle))

    def actor_tick(self, *, cycle: int):
        """
        Actor Tick
        :param cycle: cycle
        :return:
        """
        try:
            if not self.recovered:
                self.logger.warning("Tick for an actor that has not completed recovery")
                return
            if self.first_tick:
                current_cycle = cycle
            else:
                current_cycle = self.current_cycle + 1

            while current_cycle <= cycle:
                self.logger.info("actor_tick: {} start".format(current_cycle))
                self.current_cycle = current_cycle
                begin = time.time()
                self.policy.prepare(cycle=self.current_cycle)
                self.logger.info(f"POLICY TIME: {time.time() - begin:.0f}")

                if self.first_tick:
                    self.reset()

                begin = time.time()
                self.tick_handler()
                self.logger.info(f"ACTOR TICK TIME: {time.time() - begin:.0f}")
                begin = time.time()
                self.policy.finish(cycle=self.current_cycle)
                self.logger.info(f"POLICY FINISH TIME: {time.time() - begin:.0f}")

                begin = time.time()
                self.wrapper.tick()
                self.logger.info(f"KERNEL TIME: {time.time() - begin:.0f}")

                self.first_tick = False
                self.logger.info("actor_tick: {} end".format(current_cycle))
                current_cycle += 1
        except Exception as e:
            self.logger.debug(traceback.format_exc())
            raise e

    def get_actor_clock(self) -> ActorClock:
        return self.clock

    def get_client_slices(self) -> List[ABCSlice]:
        return self.wrapper.get_client_slices()

    def get_current_cycle(self) -> int:
        return self.current_cycle

    def get_description(self) -> str:
        return self.description

    def get_guid(self) -> ID or None:
        if self.identity is not None:
            return self.identity.get_guid()
        return None

    def get_identity(self) -> AuthToken:
        return self.identity

    def get_inventory_slices(self) -> List[ABCSlice]:
        """
        Get inventory slices
        @return inventory slices
        """
        return self.wrapper.get_inventory_slices()

    def get_logger(self):
        return self.logger

    def get_name(self) -> str:
        return self.name

    def get_policy(self) -> ABCPolicy:
        return self.policy

    def get_delegation(self, *, did: str) -> ABCDelegation:
        return self.wrapper.get_delegation(did=did)

    def get_reservation(self, *, rid: ID) -> ABCReservationMixin:
        return self.wrapper.get_reservation(rid=rid)

    def get_reservations(self, *, slice_id: ID) -> List[ABCReservationMixin]:
        return self.wrapper.get_reservations(slice_id=slice_id)

    def get_plugin(self):
        return self.plugin

    def get_slice(self, *, slice_id: ID) -> ABCSlice:
        return self.wrapper.get_slice(slice_id=slice_id)

    def get_slices(self):
        return self.wrapper.get_slices()

    def get_type(self) -> ActorType:
        return self.type

    def initialize(self, *, config: ActorConfig):
        from fabric_cf.actor.core.container.globals import GlobalsSingleton

        if not self.initialized:
            if self.identity is None or self.plugin is None or self.policy is None:
                raise ActorException(f"The actor is not properly created: identity: {self.identity} "
                                     f"plugin: {self.plugin} policy: {self.policy}")

            if self.name is None:
                self.name = self.identity.get_name()

            if self.name is None:
                raise ActorException("The actor is not properly created: no name")

            if self.clock is None:
                self.clock = GlobalsSingleton.get().get_container().get_actor_clock()

            if self.clock is None:
                raise ActorException("The actor is not properly created: no clock")

            if self.logger is None:
                self.logger = GlobalsSingleton.get().get_logger()

            self.plugin.set_actor(actor=self)
            self.plugin.set_logger(logger=self.logger)
            self.plugin.initialize()

            self.policy.set_actor(actor=self)
            self.policy.initialize(config=config)
            self.policy.set_logger(logger=self.logger)

            self.wrapper = KernelWrapper(actor=self, plugin=self.plugin, policy=self.policy)

            self.current_cycle = -1

            for x in self.SUPPORTED_EVENTS:
                self.event_processors[x] = EventProcessor(name=str(x), logger=self.logger)
            self.setup_message_service()

            self.initialized = True

    def is_recovered(self) -> bool:
        return self.recovered

    def is_stopped(self) -> bool:
        return self.stopped

    def query(self, *, query: dict = None, caller: AuthToken = None, actor_proxy: ABCActorProxy = None,
              handler: ABCQueryResponseHandler = None) -> dict or None:
        """
        Query an actor
        @param query query
        @param caller caller
        @param actor_proxy actor proxy
        @param handler response handler
        """
        if actor_proxy is None and handler is None:
            return self.wrapper.query(properties=query, caller=caller)
        else:
            callback = Proxy.get_callback(actor=self, protocol=actor_proxy.get_type())
            RPCManagerSingleton.get().query(actor=self, remote_actor=actor_proxy, callback=callback, query=query,
                                            handler=handler)
            return None

    def recover(self):
        """
        Recover
        """
        self.logger.info("Starting recovery")
        self.recovery_starting()
        self.logger.debug("Recovering inventory slices")

        inventory_slices = self.plugin.get_database().get_slices(slc_type=[SliceTypes.InventorySlice])
        self.logger.debug("Found {} inventory slices".format(len(inventory_slices)))
        self.recover_slices(slices=inventory_slices)
        self.logger.debug("Recovery of inventory slices complete")

        self.logger.debug("Recovering client slices")
        client_slices = self.plugin.get_database().get_slices(slc_type=[SliceTypes.ClientSlice,
                                                                        SliceTypes.BrokerClientSlice],
                                                              states=[SliceState.Configuring.value,
                                                                     SliceState.Nascent.value,
                                                                     SliceState.StableOK.value,
                                                                     SliceState.StableError.value,
                                                                     SliceState.Modifying.value,
                                                                     SliceState.ModifyOK.value,
                                                                     SliceState.ModifyError.value])
        self.logger.debug("Found {} client slices".format(len(client_slices)))
        self.recover_slices(slices=client_slices)
        self.logger.debug("Recovery of client slices complete")

        self.recovered = True

        self.recovery_ended()
        self.logger.info("Recovery complete")

    def recovery_starting(self):
        """
        Recovery starting
        """
        self.plugin.recovery_starting()
        self.policy.recovery_starting()

    def recovery_ended(self):
        """
        Recovery ended
        """
        self.plugin.recovery_ended()
        self.policy.recovery_ended()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        if GlobalsSingleton.get().can_reload_model():
            GlobalsSingleton.get().delete_reload_model_state_file()

    def recover_slices(self, *, slices: List[ABCSlice]):
        """
        Recover slices
        @param slices slices
        """
        for s in slices:
            try:
                self.recover_slice(slice_obj=s)
            except Exception as e:
                self.logger.error(traceback.format_exc())
                self.logger.error("Error in recoverSlice for property list {}".format(e))
                if s.is_inventory():
                    raise e

    def recover_broker_slice(self, *, slice_obj: ABCSlice):
        """
        Recover broker slice at the AM, do the following if the model.reload file is detected
        - Close the existing delegations
        - Create the new delegations from the reloaded ARM
        - Add the delegations to the Broker Slice

        @param slice_obj Slice object
        """
        if self.get_type() != ActorType.Authority:
            return False

        if not slice_obj.is_broker_client():
            return False

        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        if not GlobalsSingleton.get().can_reload_model():
            return False

        self.logger.info(f"Closing old delegations and adding new delegations to the slice: {slice_obj}!")
        delegation_guids = {}

        try:
            delegations = self.plugin.get_database().get_delegations(slice_id=str(slice_obj.get_slice_id()))
        except Exception as e:
            self.logger.error(e)
            raise ActorException(f"Could not fetch delegations records for slice {slice_obj} from database")

        for d in delegations:
            # Skip closed delegation
            if d.is_closed():
                self.logger.info(f"Skipping closed delegation: {d}!")
                continue
            # Grab Delegation Name and Delegation Id
            delegation_guids[d.get_delegation_name()] = d.get_delegation_id()

        # Generate ADMs and use existing delegation Id
        adms = self.policy.aggregate_resource_model.generate_adms(delegation_guids=delegation_guids)

        # Update the delegations with the new graph
        for d in delegations:
            # Skip closed delegation
            if d.is_closed():
                continue

            new_delegation_graph = adms.get(d.get_delegation_name())
            if new_delegation_graph.get_graph_id() == d.get_delegation_id():
                self.logger.info(f"Updating graph for delegation: {d}")
                d.set_graph(graph=new_delegation_graph)
            self.plugin.get_database().update_delegation(delegation=d)

    def recover_inventory_slice(self, *, slice_obj: ABCSlice) -> bool:
        """
        Check and Reload ARM for an inventory slice for an AM
        @param slice_obj slice object
        @return True if ARM was reloaded; otherwise False
        """
        if self.get_type() != ActorType.Authority:
            return False

        if not slice_obj.is_inventory():
            return False

        # Check and Reload ARM if needed
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        arm_graph = GlobalsSingleton.get().check_and_reload_model(graph_id=slice_obj.get_graph_id())
        if arm_graph is not None:
            slice_obj.set_graph(graph=arm_graph)

        return arm_graph is not None

    def recover_slice(self, *, slice_obj: ABCSlice):
        """
        Recover slice
        @param slice_obj slice_obj
        """
        slice_id = slice_obj.get_slice_id()

        if self.get_slice(slice_id=slice_id) is not None:
            self.logger.debug("Found slice_id: {} slice:{}".format(slice_id, slice_obj))
        else:
            self.logger.info("Recovering slice: {}".format(slice_id))
            self.recover_inventory_slice(slice_obj=slice_obj)

            self.recover_broker_slice(slice_obj=slice_obj)

            self.logger.debug("Informing the plugin about the slice")
            self.plugin.revisit(slice_obj=slice_obj)

            self.logger.debug("Registering slice: {}".format(slice_id))
            self.re_register_slice(slice_object=slice_obj)

            self.logger.debug("Recovering reservations in slice: {}".format(slice_id))
            self.recover_reservations(slice_obj=slice_obj)

            self.logger.debug("Recovering delegations in slice: {}".format(slice_id))
            self.recover_delegations(slice_obj=slice_obj)

            self.logger.info("Recovery of slice {} complete".format(slice_id))

    def recover_reservations(self, *, slice_obj: ABCSlice):
        """
        Recover reservations
        @param slice_obj slice object
        """
        self.logger.info(
            "Starting to recover reservations in slice {}({})".format(slice_obj.get_name(), slice_obj.get_slice_id()))
        try:
            reservations = self.plugin.get_database().get_reservations(slice_id=slice_obj.get_slice_id())
        except Exception as e:
            self.logger.error(e)
            raise ActorException(
                "Could not fetch reservation records for slice {}({}) from database".format(slice_obj.get_name(),
                                                                                            slice_obj.get_slice_id()))

        self.logger.debug("There are {} reservations(s) in slice".format(len(reservations)))

        for r in reservations:
            try:
                self.recover_reservation(r=r, slice_obj=slice_obj)
            except Exception as e:
                self.logger.error("Unexpected error while recovering reservation {}".format(e))

        self.logger.info("Recovery for reservations in slice {} completed".format(slice_obj))

    def recover_reservation(self, *, r: ABCReservationMixin, slice_obj: ABCSlice):
        """
        Recover reservation
        @param r reservation
        @param slice_obj slice object
        """
        try:
            r.restore(actor=self, slice_obj=slice_obj)

            self.logger.info(
                "Found reservation # {} in state {}".format(r.get_reservation_id(), r.get_reservation_state()))
            if r.is_closed():
                self.logger.info("Reservation #{} is closed. Nothing to recover.".format(r.get_reservation_id()))
                return

            self.logger.info("Recovering reservation #{}".format(r.get_reservation_id()))
            self.logger.debug("Recovering reservation object r={}".format(r))

            self.logger.debug("Registering the reservation with the actor")
            self.re_register(reservation=r)

            self.logger.info(r)

            self.logger.debug("Revisiting with the Plugin")

            self.plugin.revisit(reservation=r)

            self.logger.info(r)

            self.logger.debug("Revisiting with the actor policy")
            self.policy.revisit(reservation=r)

            self.logger.info("Recovered reservation #{}".format(r.get_reservation_id()))
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Exception occurred in recovering reservation e={}".format(e))
            raise ActorException("Could not recover Reservation #{}".format(r))

    def recover_delegations(self, *, slice_obj: ABCSlice):
        """
        Recover delegations for a slice
        @param slice_obj slice object
        """
        self.logger.info(
            "Starting to recover delegations in slice {}({})".format(slice_obj.get_name(), slice_obj.get_slice_id()))
        try:
            delegations = self.plugin.get_database().get_delegations(slice_id=str(slice_obj.get_slice_id()))
        except Exception as e:
            self.logger.error(e)
            raise ActorException(
                "Could not fetch delegations records for slice {}({}) from database".format(slice_obj.get_name(),
                                                                                            slice_obj.get_slice_id()))

        self.logger.debug("There are {} delegations(s) in slice".format(len(delegations)))

        for d in delegations:
            try:
                self.logger.info("Delegation has properties: {}".format(d))
                self.recover_delegation(d=d, slice_obj=slice_obj)
            except Exception as e:
                self.logger.error("Unexpected error while recovering delegation {}".format(e))

        self.logger.info("Recovery for delegations in slice {} completed".format(slice_obj))

    def recover_delegation(self, *, d: ABCDelegation, slice_obj: ABCSlice):
        """
        Recover delegation
        @param d delegation
        @param slice_obj slice object
        """
        try:

            d.restore(actor=self, slice_obj=slice_obj)

            self.logger.info(
                "Found delegation # {} in state {}".format(d.get_delegation_id(), d.get_state_name()))
            if d.is_closed():
                self.logger.info("Delegation #{} is closed. Nothing to recover.".format(d.get_delegation_id()))
                return

            self.logger.info("Recovering delegation #{}".format(d.get_delegation_id()))
            self.logger.debug("Recovering delegation object d={}".format(d))

            self.logger.debug("Registering the delegation with the actor")
            self.re_register_delegation(delegation=d)

            self.logger.info(d)

            self.logger.debug("Revisiting with the Plugin")

            self.plugin.revisit(delegation=d)

            self.logger.info(d)

            self.logger.debug("Revisiting with the actor policy")
            self.policy.revisit_delegation(delegation=d)

            self.logger.info("Recovered delegation #{}".format(d.get_delegation_id()))
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Exception occurred in recovering delegation e={}".format(e))
            raise ActorException("Could not recover delegation #{}".format(d))

    def register(self, *, reservation: ABCReservationMixin):
        self.wrapper.register_reservation(reservation=reservation)

    def register_slice(self, *, slice_object: ABCSlice):
        self.wrapper.register_slice(slice_object=slice_object)

    def modify_slice(self, *, slice_object: ABCSlice):
        self.wrapper.modify_slice(slice_object=slice_object)

    def delete_slice(self, *, slice_id: ID):
        self.wrapper.delete_slice(slice_id=slice_id)

    def register_delegation(self, *, delegation: ABCDelegation):
        self.wrapper.register_delegation(delegation=delegation)

    def remove_delegation(self, *, did: str):
        self.wrapper.remove_delegation(did=did)

    def remove_reservation(self, *, reservation: ABCReservationMixin = None, rid: ID = None):
        if reservation is not None:
            self.wrapper.remove_reservation(rid=reservation.get_reservation_id())

        if rid is not None:
            self.wrapper.remove_reservation(rid=rid)

    def remove_slice(self, *, slice_object: ABCSlice):
        self.wrapper.remove_slice(slice_id=slice_object.get_slice_id())

    def remove_slice_by_slice_id(self, *, slice_id: ID):
        self.wrapper.remove_slice(slice_id=slice_id)

    def modify_accept(self, *, slice_id: ID):
        self.wrapper.modify_accept(slice_id=slice_id)

    def re_register_delegation(self, *, delegation: ABCDelegation):
        self.wrapper.re_register_delegation(delegation=delegation)

    def re_register(self, *, reservation: ABCReservationMixin):
        self.wrapper.re_register_reservation(reservation=reservation)

    def re_register_slice(self, *, slice_object: ABCSlice):
        self.wrapper.re_register_slice(slice_object=slice_object)

    def issue_delayed(self):
        """
        Issues delayed operations
        """
        assert self.recovered
        self.close_reservations(reservations=self.closing)
        self.closing.clear()

    def reset(self):
        """
        Reset an actor
        """
        self.issue_delayed()
        self.policy.reset()

    def set_actor_clock(self, *, clock):
        """
        Set actor clock
        @param clock clock
        """
        self.clock = clock

    def set_description(self, *, description: str):
        """
        Set description
        @param description description
        """
        self.description = description

    def set_identity(self, *, token: AuthToken):
        """
        Set identity
        @param token token
        """
        self.identity = token
        self.name = self.identity.get_name()
        self.guid = token.get_guid()

    def set_policy(self, *, policy):
        """
        Set policy
        @param policy policy
        """
        self.policy = policy

    def set_recovered(self, *, value: bool):
        """
        Set recovered flag
        @param value value
        """
        self.recovered = value

    def set_plugin(self, *, plugin):
        """
        Set plugin
        @param plugin
        """
        self.plugin = plugin

    def set_stopped(self, *, value: bool):
        """
        Set stopped flag
        @param value value
        """
        self.stopped = value

    def execute_on_actor_thread(self, *, runnable: ABCActorRunnable):
        """
        Execute an incoming action on actor thread
        @param runnable incoming action/operation
        """
        self.event_processors[EventType.InterActorEvent].execute_on_thread_async(runnable=runnable)

    def execute_on_actor_thread_and_wait(self, *, runnable: ABCActorRunnable):
        """
        Execute an incoming action on actor thread
        @param runnable incoming action/operation
        """
        return self.event_processors[EventType.SyncEvent].execute_on_thread_sync(runnable=runnable)

    def start(self):
        """
        Start an Actor
        """
        for x in self.event_processors.values():
            x.start()
        self.message_service.start()
        if self.plugin.get_handler_processor() is not None:
            self.plugin.get_handler_processor().start()

    def stop(self):
        """
        Stop an actor
        """
        self.stopped = True
        self.message_service.stop()

        for x in self.event_processors.values():
            x.stop()

        if self.plugin.get_handler_processor() is not None:
            self.plugin.get_handler_processor().shutdown()

    def tick_handler(self):
        """
        Tick handler
        """

    def handle_failed_rpc(self, *, rid: ID, rpc: FailedRPC):
        """
        Handler failed rpc
        """
        self.wrapper.process_failed_rpc(rid=rid, rpc=rpc)

    def __str__(self):
        return "actor: [{}/{}]".format(self.name, self.guid)

    def unregister(self, *, reservation: ABCReservationMixin, rid: ID):
        """
        Unregister reservation
        @param reservation reservation
        @param rid reservation id
        """
        if reservation is not None:
            self.wrapper.unregister_reservation(rid=reservation.get_reservation_id())

        if rid is not None:
            self.wrapper.unregister_reservation(rid=rid)

    def unregister_slice(self, *, slice_object: ABCSlice):
        """
        Unregister slice
        @param slice_object slice object
        """
        self.wrapper.unregister_slice(slice_id=slice_object.get_slice_id())

    def unregister_slice_by_slice_id(self, *, slice_id: ID):
        """
        Unregister slice by slice id
        @param slice_id slice id
        """
        self.wrapper.unregister_slice(slice_id=slice_id)

    def queue_timer(self, timer: ABCTimerTask):
        """
        Queue an event on Actor timer queue
        """
        try:
            self.event_processors[EventType.InterActorEvent].enqueue(incoming=timer)
            self.logger.debug("Added timer to timer queue {}".format(timer.__class__.__name__))
        except Exception as e:
            self.logger.error(f"Failed to queue event: {timer.__class__.__name__} e: {e}")

    def queue_event(self, *, incoming: ABCActorEvent):
        """
        Queue an even on Actor Event Queue
        """
        try:
            if isinstance(incoming, TickEvent):
                self.event_processors[EventType.TickEvent].enqueue(incoming=incoming)
            else:
                self.event_processors[EventType.InterActorEvent].enqueue(incoming=incoming)
            self.logger.debug("Added event to event queue {}".format(incoming.__class__.__name__))
        except Exception as e:
            self.logger.error(f"Failed to queue event: {incoming.__class__.__name__} e: {e}")
            self.logger.error(traceback.format_exc())

    def queue_event_sync(self, *, incoming: ABCActorEvent):
        """
        Queue an even on Actor Event Queue
        """
        try:
            self.event_processors[EventType.SyncEvent].enqueue(incoming=incoming)
            self.logger.debug("Added event to event queue {}".format(incoming.__class__.__name__))
        except Exception as e:
            self.logger.error(f"Failed to queue event: {incoming.__class__.__name__} e: {e}")

    def await_no_pending_reservations(self):
        """
        Await until no pending reservations
        """
        self.wrapper.await_nothing_pending()

    def setup_message_service(self):
        """
        Set up Message Service for incoming Kafka Messages
        """
        try:
            # Kafka Proxy Service object
            module_name = self.get_kafka_service_module()
            class_name = self.get_kafka_service_class()
            kafka_service = ReflectionUtils.create_instance_with_params(module_name=module_name,
                                                                        class_name=class_name)(actor=self)

            # Kafka Management Service object
            module_name = self.get_mgmt_kafka_service_module()
            class_name = self.get_mgmt_kafka_service_class()
            kafka_mgmt_service = ReflectionUtils.create_instance_with_params(module_name=module_name,
                                                                             class_name=class_name)()
            kafka_mgmt_service.set_logger(logger=self.logger)

            # Incoming Message Service
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            config = GlobalsSingleton.get().get_config()
            topic = config.get_actor_config().get_kafka_topic()
            if "," in topic:
                topics = topic.split(',')
            else:
                topics = [topic]
            consumer_conf = GlobalsSingleton.get().get_kafka_config_consumer()
            self.message_service = MessageService(kafka_service=kafka_service, kafka_mgmt_service=kafka_mgmt_service,
                                                  consumer_conf=consumer_conf,
                                                  key_schema_location=GlobalsSingleton.get().get_config().get_kafka_key_schema_location(),
                                                  value_schema_location=GlobalsSingleton.get().get_config().get_kafka_value_schema_location(),
                                                  topics=topics, logger=self.logger)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Failed to setup message service e={}".format(e))
            raise e

    def set_logger(self, logger):
        self.logger = logger
        if self.policy is not None:
            self.policy.set_logger(logger=logger)
        if self.plugin is not None:
            self.plugin.set_logger(logger=logger)

    def load_model(self, *, graph_id: str):
        return

    def update_maintenance_mode(self, *, properties: Dict[str, str], sites: List[Site] = None):
        self.wrapper.update_maintenance_mode(properties=properties, sites=sites)



# coding: utf-8

# flake8: noqa
from __future__ import absolute_import
# import models into model package
from fabric_cf.orchestrator.swagger_server.models.poa import Poa
from fabric_cf.orchestrator.swagger_server.models.poa_data import PoaData
from fabric_cf.orchestrator.swagger_server.models.poa_post import PoaPost
from fabric_cf.orchestrator.swagger_server.models.poa_post_data import PoaPostData
from fabric_cf.orchestrator.swagger_server.models.poa_post_data_keys import PoaPostDataKeys
from fabric_cf.orchestrator.swagger_server.models.poa_post_data_vcpu_cpu_map import PoaPostDataVcpuCpuMap
from fabric_cf.orchestrator.swagger_server.models.resource import Resource
from fabric_cf.orchestrator.swagger_server.models.resources import Resources
from fabric_cf.orchestrator.swagger_server.models.slice import Slice
from fabric_cf.orchestrator.swagger_server.models.slice_details import SliceDetails
from fabric_cf.orchestrator.swagger_server.models.slices import Slices
from fabric_cf.orchestrator.swagger_server.models.slices_post import SlicesPost
from fabric_cf.orchestrator.swagger_server.models.sliver import Sliver
from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers
from fabric_cf.orchestrator.swagger_server.models.status200_ok_no_content import Status200OkNoContent
from fabric_cf.orchestrator.swagger_server.models.status200_ok_no_content_data import Status200OkNoContentData
from fabric_cf.orchestrator.swagger_server.models.status200_ok_paginated import Status200OkPaginated
from fabric_cf.orchestrator.swagger_server.models.status200_ok_single import Status200OkSingle
from fabric_cf.orchestrator.swagger_server.models.status400_bad_request import Status400BadRequest
from fabric_cf.orchestrator.swagger_server.models.status400_bad_request_errors import Status400BadRequestErrors
from fabric_cf.orchestrator.swagger_server.models.status401_unauthorized import Status401Unauthorized
from fabric_cf.orchestrator.swagger_server.models.status401_unauthorized_errors import Status401UnauthorizedErrors
from fabric_cf.orchestrator.swagger_server.models.status403_forbidden import Status403Forbidden
from fabric_cf.orchestrator.swagger_server.models.status403_forbidden_errors import Status403ForbiddenErrors
from fabric_cf.orchestrator.swagger_server.models.status404_not_found import Status404NotFound
from fabric_cf.orchestrator.swagger_server.models.status404_not_found_errors import Status404NotFoundErrors
from fabric_cf.orchestrator.swagger_server.models.status500_internal_server_error import Status500InternalServerError
from fabric_cf.orchestrator.swagger_server.models.status500_internal_server_error_errors import Status500InternalServerErrorErrors
from fabric_cf.orchestrator.swagger_server.models.version import Version
from fabric_cf.orchestrator.swagger_server.models.version_data import VersionData



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import json
import traceback
from datetime import datetime

from fabric_mb.message_bus.messages.add_slice_avro import AddSliceAvro
from fabric_mb.message_bus.messages.close_delegations_avro import CloseDelegationsAvro
from fabric_mb.message_bus.messages.close_reservations_avro import CloseReservationsAvro
from fabric_mb.message_bus.messages.get_delegations_avro import GetDelegationsAvro
from fabric_mb.message_bus.messages.get_reservations_request_avro import GetReservationsRequestAvro
from fabric_mb.message_bus.messages.get_reservations_state_request_avro import GetReservationsStateRequestAvro
from fabric_mb.message_bus.messages.get_sites_request_avro import GetSitesRequestAvro
from fabric_mb.message_bus.messages.get_slices_request_avro import GetSlicesRequestAvro
from fabric_mb.message_bus.messages.maintenance_request_avro import MaintenanceRequestAvro
from fabric_mb.message_bus.messages.remove_delegation_avro import RemoveDelegationAvro
from fabric_mb.message_bus.messages.remove_reservation_avro import RemoveReservationAvro
from fabric_mb.message_bus.messages.remove_slice_avro import RemoveSliceAvro
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fabric_mb.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fabric_mb.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric_mb.message_bus.messages.result_reservation_state_avro import ResultReservationStateAvro
from fabric_mb.message_bus.messages.result_sites_avro import ResultSitesAvro
from fabric_mb.message_bus.messages.result_slice_avro import ResultSliceAvro
from fabric_mb.message_bus.messages.result_string_avro import ResultStringAvro
from fabric_mb.message_bus.messages.update_reservation_avro import UpdateReservationAvro
from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro
from fabric_mb.message_bus.messages.update_slice_avro import UpdateSliceAvro
from fim.slivers.base_sliver import BaseSliver

from fabric_cf.actor.core.common.constants import ErrorCodes
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
from fabric_cf.actor.core.manage.management_object import ManagementObject
from fabric_cf.actor.core.manage.kafka.services.kafka_service import KafkaService
from fabric_cf.actor.core.proxies.kafka.translate import Translate
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.utils import translate_avro_message_type_pdp_action_id
from fabric_cf.actor.security.pdp_auth import ActionId


class KafkaActorService(KafkaService):
    def authorize_request(self, *, id_token: str, message_name: str, resource: BaseSliver = None,
                          lease_end_time: datetime = None) -> ResultStringAvro or None:
        """
        Authorize request
        :param id_token:
        :param message_name:
        :param resource:
        :param lease_end_time:
        :return:
        """
        result = None
        action_id = translate_avro_message_type_pdp_action_id(message_name=message_name)

        # Internal messages do not require authorization
        if action_id is ActionId.noop:
            return result

        try:
            from fabric_cf.actor.security.access_checker import AccessChecker
            AccessChecker.check_access(action_id=action_id, token=id_token, logger=self.logger,
                                       resource=resource, lease_end_time=lease_end_time)
        except Exception as e:
            self.logger.exception(e)
            result = ResultStringAvro()
            status = ResultAvro()
            status.status = ErrorCodes.unauthorized
            status.message = str(e)
            result.status = status
        return result

    def process(self, *, message: AbcMessageAvro):
        callback_topic = message.get_callback_topic()
        result = None

        self.logger.debug("Processing message: {}".format(message.get_message_name()))

        result = self.authorize_request(id_token=message.get_id_token(), message_name=message.get_message_name())

        # If authorization failed, return the result
        if result is None:
            if message.get_message_name() == AbcMessageAvro.get_slices_request:
                result = self.get_slices(request=message)

            elif message.get_message_name() == AbcMessageAvro.remove_slice:
                result = self.remove_slice(request=message)

            elif message.get_message_name() == AbcMessageAvro.add_slice:
                result = self.add_slice(request=message)

            elif message.get_message_name() == AbcMessageAvro.update_slice:
                result = self.update_slice(request=message)

            elif message.get_message_name() == AbcMessageAvro.get_reservations_request:
                result = self.get_reservations(request=message)

            elif message.get_message_name() == AbcMessageAvro.get_sites_request:
                result = self.get_sites(request=message)

            elif message.get_message_name() == AbcMessageAvro.get_delegations:
                result = self.get_delegations(request=message)

            elif message.get_message_name() == AbcMessageAvro.remove_reservation:
                result = self.remove_reservation(request=message)

            elif message.get_message_name() == AbcMessageAvro.close_reservations:
                result = self.close_reservations(request=message)

            elif message.get_message_name() == AbcMessageAvro.update_reservation:
                result = self.update_reservation(request=message)

            elif message.get_message_name() == AbcMessageAvro.get_reservations_state_request:
                result = self.get_reservation_state(request=message)

            elif message.get_message_name() == AbcMessageAvro.maintenance_request:
                result = self.maintenance_request(request=message)

            else:
                self.logger.debug("Unsupported Message, discarding it!")
                return

        if callback_topic is None:
            self.logger.debug("No callback specified, ignoring the message")

        if self.producer.produce(topic=callback_topic, record=result):
            self.logger.debug("Successfully send back response: {}".format(result.to_dict()))
        else:
            self.logger.debug("Failed to send back response: {}".format(result.to_dict()))

    def get_actor_mo(self, *, guid: ID):
        try:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            return GlobalsSingleton.get().get_container().get_management_object_manager().get_management_object(key=guid)
        except Exception as e:
            raise ManageException("Invalid actor id={}! e={}".format(guid, e))

    def get_slices(self, *, request: GetSlicesRequestAvro) -> ResultSliceAvro:
        result = ResultSliceAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id
        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result
            states = SliceState.list_values_ex_closing_dead()
            if request.states is not None:
                states = []
                for x in request.states:
                    if x == SliceState.All.value:
                        states = SliceState.list_values()
                        break
                    else:
                        states.append(x)

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))
            slice_id = ID(uid=request.get_slice_id()) if request.slice_id is not None else None
            result = mo.get_slices(slice_id=slice_id, caller=auth, slice_name=request.get_slice_name(),
                                   email=request.get_email(), states=states)

        except Exception as e:
            self.logger.error(traceback.format_exc())
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def add_slice(self, *, request: AddSliceAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None and request.slice_obj is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.add_slice(slice_obj=request.slice_obj, caller=auth)
        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def update_slice(self, *, request: UpdateSliceAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None and request.slice_obj is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result.status = mo.update_slice(request.slice_obj, auth)

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def remove_slice(self, *, request: RemoveSliceAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None and request.slice_id is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result.status = mo.remove_slice(slice_id=ID(uid=request.slice_id), caller=auth)

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def get_reservations(self, *, request: GetReservationsRequestAvro) -> ResultReservationAvro:
        result = ResultReservationAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))
            slice_id = ID(uid=request.slice_id) if request.slice_id is not None else None
            rid = ID(uid=request.reservation_id) if request.reservation_id is not None else None

            result = mo.get_reservations(caller=auth, states=request.get_states(), slice_id=slice_id,
                                         rid=rid, email=request.get_email(), type=request.get_type(),
                                         site=request.get_site())

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def get_sites(self, *, request: GetSitesRequestAvro) -> ResultSitesAvro:
        result = ResultSitesAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.get_sites(caller=auth, site=request.get_site())

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def update_reservation(self, *, request: UpdateReservationAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None or request.get_reservation() is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result.status = mo.update_reservation(reservation=request.reservation_obj, caller=auth)

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def remove_reservation(self, *, request: RemoveReservationAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None or request.get_reservation_id() is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))
            result.status = mo.remove_reservation(caller=auth, rid=ID(uid=request.reservation_id))

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def close_reservations(self, *, request: CloseReservationsAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None or (request.get_slice_id() is None and request.get_reservation_id() is None):
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            if request.get_slice_id() is not None:
                result.status = mo.close_slice_reservations(caller=auth, slice_id=ID(uid=request.slice_id))
            else:
                result.status = mo.close_reservation(caller=auth, rid=ID(uid=request.reservation_id))

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def get_reservation_state(self, *, request: GetReservationsStateRequestAvro) -> ResultReservationStateAvro:
        result = ResultReservationStateAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None or request.get_reservation_ids() is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.get_reservation_state_for_reservations(caller=auth, rids=request.get_reservation_ids())

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def get_delegations(self, *, request: GetDelegationsAvro) -> ResultDelegationAvro:
        result = ResultDelegationAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            slice_id = ID(uid=request.slice_id) if request.slice_id is not None else None
            result = mo.get_delegations(caller=auth, did=request.get_delegation_id(),
                                        states=request.get_states(), slice_id=slice_id)

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def maintenance_request(self, *, request: MaintenanceRequestAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        try:
            if request.actor_guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            mo = self.get_actor_mo(guid=ID(uid=request.actor_guid))
            sites = None
            if request.get_sites() is not None:
                sites = []
                for s in request.get_sites():
                    sites.append(Translate.translate_site_from_avro(site_avro=s))
            mo.update_maintenance_mode(properties=request.get_properties(), sites=sites)

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def remove_delegation(self, *, request: RemoveDelegationAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None or request.get_delegation_id() is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))
            result.status = mo.remove_delegation(caller=auth, did=request.delegation_id)

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result

    def close_delegations(self, *, request: CloseDelegationsAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        result.message_id = request.message_id

        try:
            if request.guid is None or (request.get_slice_id() is None and request.get_delegation_id() is None):
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result.status = mo.close_delegation(caller=auth, did=request.delegation_id)

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        result.message_id = request.message_id
        return result


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import enum
from enum import Enum
from typing import Tuple, List

from fabric_cf.actor.core.common.exceptions import SliceException
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.actor.core.util.id import ID

from fabric_cf.actor.core.util.reservation_set import ReservationSet


class SliceState(Enum):
    Nascent = enum.auto()
    Configuring = enum.auto()
    StableError = enum.auto()
    StableOK = enum.auto()
    Closing = enum.auto()
    Dead = enum.auto()
    Modifying = enum.auto()
    ModifyError = enum.auto()
    ModifyOK = enum.auto()
    All = enum.auto()   # used only for querying

    def __str__(self):
        return self.name

    @classmethod
    def list_values(cls) -> List[int]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def list_names(cls) -> List[str]:
        return list(map(lambda c: c.name, cls))

    @staticmethod
    def list_values_ex_closing_dead() -> List[int]:
        result = SliceState.list_values()
        result.remove(SliceState.Closing.value)
        result.remove(SliceState.Dead.value)
        return result

    @staticmethod
    def translate_list(states: List[str]) -> List[int] or None:
        if states is None or len(states) == 0:
            return states

        incoming_states = list(map(lambda x: x.lower(), states))

        result = SliceState.list_values()

        if len(incoming_states) == 1 and incoming_states[0] == SliceState.All.name.lower():
            return result

        for s in SliceState:
            if s.name.lower() not in incoming_states:
                result.remove(s.value)

        return result

    @staticmethod
    def translate(state_name: str):
        if state_name.lower() == SliceState.Nascent.name.lower():
            return SliceState.Nascent
        elif state_name.lower() == SliceState.Configuring.name.lower():
            return SliceState.Configuring
        elif state_name.lower() == SliceState.StableOK.name.lower():
            return SliceState.StableOK
        elif state_name.lower() == SliceState.StableError.name.lower():
            return SliceState.StableError
        elif state_name.lower() == SliceState.ModifyOK.name.lower():
            return SliceState.ModifyOK
        elif state_name.lower() == SliceState.ModifyError.name.lower():
            return SliceState.ModifyError
        elif state_name.lower() == SliceState.Modifying.name.lower():
            return SliceState.Modifying
        elif state_name.lower() == SliceState.Closing.name.lower():
            return SliceState.Closing
        elif state_name.lower() == SliceState.Dead.name.lower():
            return SliceState.Dead
        else:
            return SliceState.All

    @staticmethod
    def is_dead_or_closing(*, state) -> bool:
        if state == SliceState.Dead or state == SliceState.Closing:
            return True
        return False

    @staticmethod
    def is_stable(*, state) -> bool:
        if state == SliceState.StableOK or state == SliceState.StableError:
            return True
        return False

    @staticmethod
    def is_modified(*, state) -> bool:
        if state == SliceState.ModifyOK or state == SliceState.ModifyError:
            return True
        return False


class SliceCommand(Enum):
    Create = enum.auto()
    Modify = enum.auto()
    Delete = enum.auto()
    Reevaluate = enum.auto()
    ModifyAccept = enum.auto()

    def __str__(self):
        return self.name


class SliceOperation:
    def __init__(self, command: SliceCommand, *valid_from_states):
        self.command = command
        self.valid_from_states = []
        if valid_from_states is not None:
            for state in valid_from_states:
                self.valid_from_states.append(state)

    def __str__(self):
        return self.command.name


class StateBins:
    def __init__(self):
        self.bins = {}

    def add(self, *, s: ReservationStates):
        if s not in self.bins:
            self.bins[s] = 1
        else:
            self.bins[s] += 1

    def has_state(self, *, s: ReservationStates) -> bool:
        count = self.bins.get(s, None)
        if count is not None and count > 0:
            return True
        return False

    def has_state_other_than(self, *states) -> bool:
        count = 0
        for state, value in self.bins.items():
            if value > 0:
                count += 1

        count1 = 0
        for state in states:
            if self.has_state(s=state):
                count1 += 1

        if count1 == count and count > 0:
            return False

        return True


class SliceStateMachine:
    CREATE = SliceOperation(SliceCommand.Create, SliceState.Nascent)

    MODIFY = SliceOperation(SliceCommand.Modify, SliceState.StableOK, SliceState.StableError, SliceState.Configuring)

    MODIFY_ACCEPT = SliceOperation(SliceCommand.ModifyAccept, SliceState.ModifyOK, SliceState.ModifyError,
                                   SliceState.Modifying)

    DELETE = SliceOperation(SliceCommand.Delete, SliceState.Nascent, SliceState.StableOK, SliceState.StableError,
                            SliceState.Configuring, SliceState.Modifying, SliceState.ModifyOK, SliceState.ModifyError,
                            SliceState.Dead)

    REEVALUATE = SliceOperation(SliceCommand.Reevaluate, SliceState.Nascent, SliceState.StableOK,
                                SliceState.StableError, SliceState.Configuring, SliceState.Dead, SliceState.Closing,
                                SliceState.Modifying, SliceState.ModifyError, SliceState.ModifyOK)

    def __init__(self, *, slice_id: ID):
        self.slice_guid = slice_id
        self.last_state = SliceState.Nascent
        self.state = SliceState.Nascent

    @staticmethod
    def all_failed(*, reservations: ReservationSet) -> bool:
        """
        We don't introduce a special state to flag when a slice is ALL FAILED, however this helper function helps decide
        when to GC a slice
        @return true or false
        """
        bins = StateBins()
        for r in reservations.values():
            bins.add(s=r.get_state())

        if not bins.has_state_other_than(ReservationStates.Failed):
            return True

        return False

    def transition_slice(self, *, operation: SliceOperation, reservations: ReservationSet) -> Tuple[bool, SliceState]:
        """
        Attempt to transition a slice to a new state
        @param operation slice operation
        @param reservations reservations
        @return Slice State
        @throws Exception in case of error
        """
        state_changed = False
        prev_state = self.state
        if self.state not in operation.valid_from_states:
            raise SliceException(f"Operation: {operation} cannot transition from state {self.state}")

        if operation.command == SliceCommand.Create:
            self.state = SliceState.Configuring

        elif operation.command == SliceCommand.Modify:
            self.state = SliceState.Modifying

        elif operation.command == SliceCommand.ModifyAccept:
            if self.state == SliceState.ModifyError:
                self.state = SliceState.StableError
            elif self.state == SliceState.ModifyOK:
                self.state = SliceState.StableOK

        elif operation.command == SliceCommand.Delete:
            if self.state != SliceState.Dead:
                self.state = SliceState.Closing

        elif operation.command == SliceCommand.Reevaluate:
            has_error = False
            if reservations is None or reservations.size() == 0:
                return state_changed, self.state

            bins = StateBins()
            for r in reservations.values():
                bins.add(s=r.get_state())
                if r.get_pending_state() in [ReservationPendingStates.ModifyingLease,
                                             ReservationPendingStates.ExtendingTicket,
                                             ReservationPendingStates.ExtendingLease,
                                             ReservationPendingStates.Redeeming,
                                             ReservationPendingStates.Ticketing,
                                             ReservationPendingStates.Priming,
                                             ReservationPendingStates.PrimingPoa,
                                             ReservationPendingStates.WaitingPoaResponse]:
                    bins.add(s=r.get_pending_state())
                if not has_error and r.get_error_message() is not None and len(r.get_error_message()) > 0:
                    has_error = True

            if self.state == SliceState.Nascent or self.state == SliceState.Configuring:
                if not bins.has_state_other_than(ReservationStates.Active, ReservationStates.Closed):
                    if not has_error:
                        self.state = SliceState.StableOK
                    else:
                        self.state = SliceState.StableError

                if (not bins.has_state_other_than(ReservationStates.Active, ReservationStates.Failed,
                                                  ReservationStates.Closed)) and \
                        bins.has_state(s=ReservationStates.Failed):
                    self.state = SliceState.StableError

                if not bins.has_state_other_than(ReservationStates.Closed, ReservationStates.CloseWait,
                                                 ReservationStates.Failed):
                    self.state = SliceState.Closing

            elif self.state == SliceState.Modifying:
                if not bins.has_state_other_than(ReservationStates.Active, ReservationStates.Closed):
                    if has_error:
                        self.state = SliceState.ModifyError
                    else:
                        self.state = SliceState.ModifyOK
                        self.state = SliceState.ModifyError

                if (not bins.has_state_other_than(ReservationStates.Active, ReservationStates.Failed,
                                                  ReservationStates.Closed)) and \
                        bins.has_state(s=ReservationStates.Failed):
                    self.state = SliceState.ModifyError

                if not bins.has_state_other_than(ReservationStates.Closed, ReservationStates.CloseWait,
                                                 ReservationStates.Failed):
                    self.state = SliceState.Closing

            elif self.state == SliceState.StableError or self.state == SliceState.StableOK or \
                    self.state == SliceState.ModifyError or self.state == SliceState.ModifyOK:
                if not bins.has_state_other_than(ReservationStates.Closed, ReservationStates.CloseWait,
                                                 ReservationStates.Failed):
                    self.state = SliceState.Dead

                if not bins.has_state_other_than(ReservationStates.Closed, ReservationStates.CloseWait,
                                                 ReservationPendingStates.Closing, ReservationStates.Failed):
                    self.state = SliceState.Closing

            elif self.state == SliceState.Closing and not bins.has_state_other_than(ReservationStates.CloseWait,
                                                                                    ReservationStates.Closed,
                                                                                    ReservationStates.Failed):
                self.state = SliceState.Dead
        if prev_state != self.state:
            state_changed = True
            self.last_state = self.state

        return state_changed, self.state

    def get_state(self) -> SliceState:
        return self.state

    def clear(self):
        self.state = SliceState.Nascent


import logging

import prometheus_client
received_counter = prometheus_client.Counter('Requests_Received', 'HTTP Requests', ['method', 'endpoint'])
success_counter = prometheus_client.Counter('Requests_Success', 'HTTP Success', ['method', 'endpoint'])
failure_counter = prometheus_client.Counter('Requests_Failed', 'HTTP Failures', ['method', 'endpoint'])

__API_REFERENCE__ = 'https://github.com/fabric-testbed/ControlFramework'



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import traceback
from typing import Tuple, List

from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fabric_mb.message_bus.messages.slice_avro import SliceAvro

from fabric_cf.actor.core.apis.abc_delegation import DelegationState
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
from fabric_cf.actor.core.manage.error import Error
from fabric_cf.actor.core.manage.kafka.kafka_actor import KafkaActor
from fabric_cf.actor.core.util.id import ID


class ManageHelper:
    def __init__(self, *, logger):
        self.logger = logger

    def print_result(self, *, status: ResultAvro):
        self.logger.debug("Code={}".format(status.get_code()))
        if status.message is not None:
            self.logger.debug("Message={}".format(status.message))
        if status.details is not None:
            self.logger.debug("Details={}".format(status.details))

    @staticmethod
    def get_actor(*, actor_name: str) -> KafkaActor:
        from .kafka_processor import KafkaProcessorSingleton
        actor = KafkaProcessorSingleton.get().get_mgmt_actor(name=actor_name)
        return actor

    def do_get_slices(self, *, actor_name: str, callback_topic: str, slice_id: str = None, slice_name: str = None,
                      id_token: str = None, email: str = None, states: str = None) -> Tuple[
        List[SliceAvro] or None, Error]:
        actor = self.get_actor(actor_name=actor_name)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(actor_name))
        try:
            actor.prepare(callback_topic=callback_topic)
            sid = ID(uid=slice_id) if slice_id is not None else None
            slice_states = None
            if states is not None:
                states_list = states.split(",")
                for x in states_list:
                    if slice_states is None:
                        slice_states = []
                    slice_states.append(SliceState.translate(state_name=x).value)

            result = actor.get_slices(slice_id=sid, slice_name=slice_name, email=email, states=slice_states)
            return result, actor.get_last_error()
        except Exception:
            ex_str = traceback.format_exc()
            self.logger.error(ex_str)
        return None, actor.get_last_error()

    def do_get_reservations(self, *, actor_name: str, callback_topic: str, slice_id: str = None, rid: str = None,
                            states: str = None, id_token: str = None, email: str = None, site: str = None,
                            type: str = None) -> Tuple[List[ReservationMng] or None, Error]:
        actor = self.get_actor(actor_name=actor_name)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(actor_name))
        try:
            actor.prepare(callback_topic=callback_topic)
            sid = ID(uid=slice_id) if slice_id is not None else None
            reservation_id = ID(uid=rid) if rid is not None else None
            reservation_states = None
            if states is not None:
                states_list = states.split(",")
                for x in states_list:
                    if reservation_states is None:
                        reservation_states = []
                    reservation_states.append(ReservationStates.translate(state_name=x).value)
            return actor.get_reservations(slice_id=sid, rid=reservation_id, states=reservation_states, email=email,
                                          site=site, type=type), actor.get_last_error()
        except Exception as e:
            ex_str = traceback.format_exc()
            self.logger.error(ex_str)
            traceback.print_exc()
        return None, actor.get_last_error()

    def do_get_delegations(self, *, actor_name: str, callback_topic: str, slice_id: str = None, did: str = None,
                           states: str = None, id_token: str = None) -> Tuple[List[DelegationAvro] or None, Error]:
        actor = self.get_actor(actor_name=actor_name)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(actor_name))
        try:
            actor.prepare(callback_topic=callback_topic)
            sid = None
            if slice_id is not None:
                sid = ID(uid=slice_id)
            delegation_states = None
            if states is not None:
                for x in states:
                    if delegation_states is None:
                        delegation_states = []
                    delegation_states.append(DelegationState.translate(state_name=x).value)
            return actor.get_delegations(delegation_id=did, slice_id=sid,
                                         states=delegation_states), actor.get_last_error()
        except Exception as e:
            self.logger.error(f"Exception occurred while fetching delegations: e {e}")
            self.logger.error(traceback.format_exc())
            traceback.print_exc()
        return None, actor.get_last_error()

    def do_claim_delegations(self, *, broker: str, am_guid: ID, callback_topic: str, id_token: str = None,
                             did: str = None) -> Tuple[DelegationAvro, Error]:
        """
        Claim delegations by invoking Management Actor Claim Delegations API
        @param broker broker guid
        @param am_guid am guid
        @param callback_topic callback topic
        @param id_token id token
        @param did delegation id
        @return Tuple[Delegation, Error] Delegation on success and Error in case of failure
        """
        actor = self.get_actor(actor_name=broker)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(broker))
        try:
            actor.prepare(callback_topic=callback_topic)

            dlg = actor.claim_delegations(broker=am_guid, did=did)
            return dlg, actor.get_last_error()
        except Exception as e:
            self.logger.error(f"Exception occurred e: {e}")
            self.logger.error(traceback.format_exc())

        return None, actor.get_last_error()

    def do_reclaim_delegations(self, *, broker: str, am_guid: ID, callback_topic: str, id_token: str = None,
                               did: str = None) -> Tuple[DelegationAvro, Error]:
        """
        Reclaim delegations by invoking Management Actor Claim Delegations API
        @param broker broker guid
        @param am_guid am guid
        @param callback_topic callback topic
        @param id_token id token
        @param did delegation id
        @return Tuple[Delegation, Error] Delegation on success and Error in case of failure
        """
        actor = self.get_actor(actor_name=broker)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(broker))
        try:
            actor.prepare(callback_topic=callback_topic)

            dlg = actor.reclaim_delegations(broker=am_guid, did=did)
            return dlg, actor.get_last_error()
        except Exception as e:
            self.logger.error(f"Exception occurred e: {e}")
            self.logger.error(traceback.format_exc())

        return None, actor.get_last_error()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import argparse
import logging
import traceback
from logging.handlers import RotatingFileHandler

from fim.graph.neo4j_property_graph import Neo4jGraphImporter, Neo4jPropertyGraph

from fabric_cf.actor.core.plugins.db.actor_database import ActorDatabase
from fabric_cf.actor.core.util.id import ID


class MainClass:
    """
    CLI interface to directly fetch information from postgres Database
    """
    def __init__(self, user: str, password: str, db: str, host: str = '127.0.0.1:5432'):
        self.logger = logging.getLogger("db-cli")
        file_handler = RotatingFileHandler('./db_cli.log', backupCount=5, maxBytes=50000)
        logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s [%(filename)s:%(lineno)d] [%(levelname)s] %(message)s",
                            handlers=[logging.StreamHandler(), file_handler])

        self.db = ActorDatabase(user=user, password=password, database=db, db_host=host, logger=self.logger)
        self.neo4j_config = {"url": "neo4j://0.0.0.0:9687",
                             "user": "neo4j",
                             "pass": "password",
                             "import_host_dir": "/Users/kthare10/renci/code/fabric/ControlFramework/neo4j1/imports/",
                             "import_dir": "/imports"}

    def get_slices(self, email: str = None, slice_id: str = None, slice_name: str = None):
        try:
            if slice_id is not None:
                slice_obj = self.db.get_slice(slice_id=ID(uid=slice_id))
                slice_list = [slice_obj]
            elif email is not None:
                slice_list = self.db.get_slice_by_email(email=email)
            else:
                slice_list = self.db.get_slices()

            if slice_list is not None and len(slice_list) > 0:
                for s in slice_list:
                    show_slice = slice_name is None
                    if slice_name is not None:
                        show_slice = slice_name in s.get_name()
                    if show_slice:
                        print(s)
                        print()
            else:
                print(f"No slices found: {slice_list}")
        except Exception as e:
            print(f"Exception occurred while fetching slices: {e}")
            traceback.print_exc()

    def get_slice_topology(self, graph_id: str):
        try:
            neo4j_graph_importer = Neo4jGraphImporter(url=self.neo4j_config["url"],
                                                      user=self.neo4j_config["user"],
                                                      pswd=self.neo4j_config["pass"],
                                                      import_host_dir=self.neo4j_config["import_host_dir"],
                                                      import_dir=self.neo4j_config["import_dir"],
                                                      logger=self.logger)

            slice_model = Neo4jPropertyGraph(graph_id=graph_id, importer=neo4j_graph_importer)

            print(f"Slice Model: {slice_model}")
        except Exception as e:
            print(f"Exception occurred while fetching slices: {e}")
            traceback.print_exc()

    def get_delegations(self, dlg_id: str = None):
        try:
            if dlg_id is not None:
                del_list = self.db.get_delegation(dlg_graph_id=dlg_id)
            else:
                del_list = self.db.get_delegations()
            if del_list is not None and len(del_list) > 0:
                for d in del_list:
                    print(d)
                    print()
            else:
                print(f"No delegations found: {del_list}")
        except Exception as e:
            print(f"Exception occurred while fetching delegations: {e}")
            traceback.print_exc()

    def get_sites(self):
        try:
            sites = self.db.get_sites()
            if sites is not None:
                for s in sites:
                    print(s)
        except Exception as e:
            print(f"Exception occurred while fetching sites: {e}")
            traceback.print_exc()

    def get_reservations(self, slice_id: str = None, res_id: str = None, email: str = None):
        try:
            res_list = self.db.get_reservations(slice_id=slice_id, rid=res_id, email=email)
            if res_list is not None and len(res_list) > 0:
                for r in res_list:
                    print(r)
                    print()
            else:
                print(f"No reservations found: {res_list}")
        except Exception as e:
            print(f"Exception occurred while fetching delegations: {e}")
            traceback.print_exc()

    def remove_reservation(self, sliver_id: str):
        try:
            self.db.remove_reservation(rid=ID(uid=sliver_id))
        except Exception as e:
            print(f"Exception occurred while fetching delegations: {e}")
            traceback.print_exc()

    def remove_slice(self, slice_id: str):
        try:
            self.db.remove_slice(slice_id=ID(uid=slice_id))
        except Exception as e:
            print(f"Exception occurred while fetching delegations: {e}")
            traceback.print_exc()

    def handle_command(self, args):
        if args.command == "slices":
            if args.operation is not None and args.operation == "remove":
                self.remove_slice(slice_id=args.slice_id)
            elif args.operation is not None and args.operation == "topology":
                self.get_slice_topology(graph_id=args.graph_id)
            else:
                self.get_slices(slice_id=args.slice_id, email=args.email, slice_name=args.slice_name)
        elif args.command == "slivers":
            if args.operation is not None and args.operation == "remove":
                self.remove_reservation(sliver_id=args.sliver_id)
            else:
                self.get_reservations(slice_id=args.slice_id, res_id=args.sliver_id, email=args.email)
        elif args.command == "delegations":
            self.get_delegations(dlg_id=args.delegation_id)
        elif args.command == "sites":
            self.get_sites()
        else:
            print(f"Unsupported command: {args.command}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-u", dest='user', required=True, type=str)
    parser.add_argument("-p", dest='password', required=True, type=str)
    parser.add_argument("-d", dest='database', required=True, type=str)
    parser.add_argument("-c", dest='command', required=True, type=str)
    parser.add_argument("-s", dest='slice_id', required=False, type=str)
    parser.add_argument("-g", dest='graph_id', required=False, type=str)
    parser.add_argument("-r", dest='sliver_id', required=False, type=str)
    parser.add_argument("-i", dest='delegation_id', required=False, type=str)
    parser.add_argument("-e", dest='email', required=False, type=str)
    parser.add_argument("-n", dest='slice_name', required=False, type=str)
    parser.add_argument("-o", dest='operation', required=False, type=str)
    args = parser.parse_args()

    mc = MainClass(user=args.user, password=args.password, db=args.database)
    mc.handle_command(args)




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING

from fim.graph.abc_property_graph import GraphFormat

from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation, DelegationState
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.ticket import Ticket
from fabric_cf.actor.core.apis.abc_broker_policy_mixin import ABCBrokerPolicyMixin
from fabric_cf.actor.core.core.policy import Policy
from fabric_cf.actor.core.kernel.resource_set import ResourceSet

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_broker_mixin import ABCBrokerMixin
    from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation
    from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.delegation.resource_ticket import ResourceTicket
    from fabric_cf.actor.core.apis.abc_server_reservation import ABCServerReservation
    from fabric_cf.actor.core.util.id import ID


class BrokerPolicy(Policy, ABCBrokerPolicyMixin):
    """
    Base implementation for Broker policy
    """
    def __init__(self, *, actor: ABCBrokerMixin):
        super().__init__(actor=actor)
        # Initialization status.
        self.initialized = False

    def allocate(self, *, cycle: int):
        return

    def bind(self, *, reservation: ABCBrokerReservation) -> bool:
        return False

    def bind_delegation(self, *, delegation: ABCDelegation) -> bool:
        return False

    def demand(self, *, reservation: ABCClientReservation):
        return

    def donate_delegation(self, *, delegation: ABCDelegation):
        return

    def reclaim_delegation(self, *, delegation: ABCDelegation):
        return

    def extend_broker(self, *, reservation: ABCBrokerReservation) -> bool:
        return False

    def initialize(self, *, config: ActorConfig):
        if not self.initialized:
            super().initialize(config=config)
            self.initialized = True

    def revisit(self, *, reservation: ABCReservationMixin):
        return

    def update_ticket_complete(self, *, reservation: ABCClientReservation):
        return

    def revisit_delegation(self, *, delegation: ABCDelegation):
        if delegation.get_state() == DelegationState.Delegated:
            self.bind_delegation(delegation=delegation)

    def ticket_satisfies(self, *, requested_resources: ResourceSet, actual_resources: ResourceSet,
                         requested_term: Term, actual_term: Term):
        return

    def update_delegation_complete(self, *, delegation: ABCDelegation, reclaim: bool = False):
        if not reclaim:
            self.donate_delegation(delegation=delegation)
        else:
            self.reclaim_delegation(delegation=delegation)

    def extract(self, *, source: ABCDelegation, delegation: ResourceTicket) -> ResourceSet:
        """
        Creates a new resource set using the source and the specified delegation.

        @param source source
        @param delegation delegation
        @return returns ResourceSet
        @throws Exception in case of error
        """
        extracted = ResourceSet(units=delegation.get_units(), rtype=delegation.get_resource_type())

        cset = Ticket(resource_ticket=delegation, plugin=self.actor.get_plugin(), authority=source.get_site_proxy(),
                      delegation_id=source.get_delegation_id())

        extracted.set_resources(cset=cset)
        return extracted

    def get_client_id(self, *, reservation: ABCServerReservation) -> ID:
        """
        Get Client Id
        """
        return reservation.get_client_auth_token().get_guid()

    @staticmethod
    def get_broker_query_model_query(*, level: int, bqm_format: GraphFormat = GraphFormat.GRAPHML) -> dict:
        """
        Return dictionary representing query
        :param level: Graph Level
        :param bqm_format: Graph Format
        :return dictionary representing the query
        """
        properties = {Constants.QUERY_ACTION: Constants.QUERY_ACTION_DISCOVER_BQM,
                      Constants.QUERY_DETAIL_LEVEL: str(level),
                      Constants.BROKER_QUERY_MODEL_FORMAT: str(bqm_format.value)}
        return properties

    @staticmethod
    def get_query_action(properties: dict) -> str:
        """
        Return Query action
        :param properties: properties
        :return: query action
        """
        if properties is None:
            return None
        return properties.get(Constants.QUERY_ACTION, None)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric_cf.actor.core.kernel.incoming_rpc import IncomingRPC
from fabric_cf.actor.core.kernel.poa import Poa

if TYPE_CHECKING:
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
    from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
    from fabric_cf.actor.security.auth_token import AuthToken


class IncomingPoaRPC(IncomingRPC):
    """
    Represents Incoming RPC message carrying a reservation
    """
    def __init__(self, *, message_id: ID, request_type: RPCRequestType, callback: ABCCallbackProxy = None,
                 caller: AuthToken = None, poa: Poa = None):
        super().__init__(message_id=message_id, request_type=request_type, callback=callback, caller=caller)
        self.poa = poa

    def get(self):
        """
        Get POA
        @return POA
        """
        return self.poa

    def __str__(self):
        return f"{self.poa.poa_id}"

    def get_update_data(self):
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import json
import logging
from datetime import datetime, timedelta
from enum import Enum
from typing import List

import requests
from fim.slivers.base_sliver import BaseSliver
from fim.user.topology import ExperimentTopology
from fss_utils.jwt_validate import JWTValidator

from fim.authz.attribute_collector import ResourceAuthZAttributes

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.security.fabric_token import FabricToken


class PdpAuthException(Exception):
    """
    PDP Auth Exception
    """


class ResourceType(Enum):
    """
    Resource Type Enumeration
    """
    user = 1
    slice = 2
    sliver = 3
    resources = 4
    delegation = 5


class ActionId(Enum):
    """
    Action Id Enumeration
    """
    noop = 0
    query = 1
    status = 2
    create = 3
    redeem = 4
    delete = 5
    modify = 6
    POA = 7
    renew = 8
    demand = 9
    update = 10
    close = 11
    claim = 12
    reclaim = 13
    ticket = 14
    extend = 15
    relinquish = 16
    accept = 17
    poa = 18

    def __str__(self):
        return self.name


class PdpAuth:
    """
    Responsible for Authorization against PDP
    """

    def __init__(self, *, config: dict, logger=None):
        self.config = config
        self.logger = logger

    @staticmethod
    def _headers() -> dict:
        """
        Returns Headers for REST APIs
        """
        headers = {
            "accept": "application/json",
            "Content-Type": "application/json"
        }
        return headers

    def build_pdp_request(self, *, email: str, project: str, tags: List[str], action_id: ActionId,
                          resource: BaseSliver or ExperimentTopology, lease_end_time: datetime) -> dict:
        """
        Build PDP Request
        @param email email
        @param project project
        @param tags project tags
        @param action_id Action id
        @param resource: sliver of any type or slice (ExperimentTopology)
        @param lease_end_time lease end time
        @return PDP request
        """
        if project is None or email is None:
            raise PdpAuthException("No project found in fabric token")

        attrs = ResourceAuthZAttributes()
        # collect all resource attributes
        if resource is not None:
            attrs.collect_resource_attributes(source=resource)

        # additional attributes (slice end date in datetime format)
        if lease_end_time:
            attrs.set_lifetime(lease_end_time)

        # next we need to set the owner of the resource and their projects
        # generally only the id is needed. If action is create, it's not needed at all

        attrs.set_subject_attributes(subject_id=email, project=project, project_tag=tags)

        attrs.set_resource_subject_and_project(subject_id=email, project=project)

        # finally action
        # action can be any string matching ActionId enum
        attrs.set_action(action_id.name)

        # now you can produce the json
        request_json = attrs.transform_to_pdp_request(as_json=False)

        return request_json

    def check_access(self, *, email: str, project: str, tags: List[str], action_id: ActionId,
                     resource: BaseSliver or ExperimentTopology, lease_end_time: datetime):
        """
        Check Access
        @param email email
        @param project project
        @param tags project tags
        @param action_id action id
        @param resource sliver (of any type) or slice (ExperimentTopology)
        @param lease_end_time lease end time
        @raises PdpAuthException in case of denied access or failure
        """
        if not self.config['enable']:
            self.logger.debug("Skipping PDP Authorization check as configured")
            return

        pdp_request = self.build_pdp_request(email=email, project=project, tags=tags, action_id=action_id,
                                             resource=resource, lease_end_time=lease_end_time)

        self.logger.debug("PDP Auth Request: {}".format(pdp_request))

        # send request to PDP
        response = requests.post(url=self.config['url'], headers=self._headers(), json=pdp_request)
        response_json = response.json()

        if response.status_code == 200 and response_json["Response"][0]["Decision"] == "Permit":
            self.logger.debug("PDP response: {}".format(response_json))
        else:
            self.logger.error("PDP response: {}".format(response_json))
            msg = response_json["Response"][0]["AssociatedAdvice"][0]["AttributeAssignment"][0]["Value"]
            raise PdpAuthException(f"PDP Authorization check failed - {msg}")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import PluginException
from fabric_cf.actor.core.common.resource_config import ResourceConfigBuilder

if TYPE_CHECKING:
    from fabric_cf.actor.core.plugins.base_plugin import BasePlugin
    from fabric_cf.actor.core.plugins.handlers.config_token import ConfigToken
    from fabric_cf.actor.core.plugins.handlers.configuration_mapping import ConfigurationMapping


class HandlerProcessor:
    def __init__(self):
        self.plugin = None
        self.logger = None
        self.initialized = False
        self.config_mappings = {}
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['plugin']
        del state['initialized']
        del state['lock']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.plugin = None
        self.initialized = False
        self.lock = threading.Lock()

    def initialize(self, *, config: ActorConfig):
        if not self.initialized:
            if self.plugin is None:
                raise PluginException(Constants.NOT_SPECIFIED_PREFIX.format("plugin"))
            self.logger = self.plugin.get_logger()
            self.load_config_mappings_from_db()
            self.load_config_mappings(config=config)
            self.initialized = True

    def load_config_mappings_from_db(self):
        try:
            self.lock.acquire()
            mappings = self.plugin.get_database().get_config_mappings()
            for m in mappings:
                self.config_mappings[m.get_key()] = m
        finally:
            self.lock.release()

    def load_config_mappings(self, *, config: ActorConfig):
        if config is None:
            return
        resource_configs = ResourceConfigBuilder.build_resource_config(config=config)
        if resource_configs is not None:
            for rtype, rc in resource_configs.items():
                if str(rtype) not in self.config_mappings:
                    self.logger.debug(f"Adding handler for type: {rtype} handler: {type(rc)}")
                    config_map = ResourceConfigBuilder.build_config_mapping(resource_config=rc)
                    self.add_config_mapping(mapping=config_map)
                else:
                    self.logger.debug(f"Handler for type: {rtype} exists")

    def add_config_mapping(self, *, mapping: ConfigurationMapping):
        try:
            self.lock.acquire()
            self.config_mappings[mapping.get_key()] = mapping
            self.plugin.get_database().add_config_mapping(key=mapping.get_key(), config_mapping=mapping)
        finally:
            self.lock.release()

    def create(self, unit: ConfigToken):
        self.logger.debug("Executing Create")

        result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_CREATE,
                  Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_OK,
                  Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}

        self.plugin.configuration_complete(token=unit, properties=result)
        self.logger.debug("Executing Create completed")

    def delete(self, unit: ConfigToken):
        self.logger.debug("Executing Delete")

        result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_DELETE,
                  Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_OK,
                  Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}

        self.plugin.configuration_complete(token=unit, properties=result)
        self.logger.debug("Executing Delete completed")

    def modify(self, unit: ConfigToken):
        self.logger.debug("Executing Modify")

        result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_MODIFY,
                  Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_OK,
                  Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}

        self.plugin.configuration_complete(token=unit, properties=result)
        self.logger.debug("Executing Modify completed")

    def poa(self, unit: ConfigToken, data: dict):
        self.logger.debug("Executing POA")

        result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_POA,
                  Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_OK,
                  Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}

        self.plugin.configuration_complete(token=unit, properties=result)
        self.logger.debug("Executing POA completed")

    def set_logger(self, *, logger):
        self.logger = logger

    def set_plugin(self, *, plugin: BasePlugin):
        self.plugin = plugin

    @staticmethod
    def get_action_sequence_number(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.INVALID_ARGUMENT)

        if Constants.PROPERTY_ACTION_SEQUENCE_NUMBER in properties:
            return int(properties[Constants.PROPERTY_ACTION_SEQUENCE_NUMBER])

        raise PluginException("Action Sequence Number not found")

    @staticmethod
    def get_result_code(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.INVALID_ARGUMENT)

        if Constants.PROPERTY_TARGET_RESULT_CODE in properties:
            return int(properties[Constants.PROPERTY_TARGET_RESULT_CODE])

        raise PluginException("Target Result code not found")

    @staticmethod
    def get_result_code_message(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.INVALID_ARGUMENT)

        if Constants.PROPERTY_TARGET_RESULT_CODE_MESSAGE in properties:
            return properties[Constants.PROPERTY_TARGET_RESULT_CODE_MESSAGE]

        return None

    @staticmethod
    def get_exception_message(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.INVALID_ARGUMENT)

        if Constants.PROPERTY_EXCEPTION_MESSAGE in properties:
            return properties[Constants.PROPERTY_EXCEPTION_MESSAGE]

        return None

    def shutdown(self):
        return

    def start(self):
        return

    def clean_restart(self):
        return
